#!/bin/bash

if [ -t 1 ]; then
    ESCAPE_RESET="\E[0m"
    ESCAPE_BLUE="\E[34m"
    ESCAPE_UNDERLINE="\E[4m"
fi


SCRIPT_NAME=`basename $0`

SSH_OPTIONS="-2 -C -c aes256-ctr -o KexAlgorithms=curve25519-sha256@libssh.org,diffie-hellman-group-exchange-sha256 -o MACs=hmac-sha2-512 -o ConnectTimeout=5 -o LogLevel=error -T"

DEBUG=0
HELP=0
LOCAL=1
REMOTE=0


while getopts "DhLr" OPT; do
    case $OPT in
        D) DEBUG=1 ;;
        h) HELP=1 ;;
        L) LOCAL=0 ;;
        r) REMOTE=1 ;;
        *) exit 1 ;;
    esac
done
shift $((OPTIND-1))


if [[ HELP -ne 0 ]]; then
    echo
    echo    "  SYNOPSIS"
    echo -e "  $SCRIPT_NAME [${ESCAPE_UNDERLINE}-D${ESCAPE_RESET}] [${ESCAPE_UNDERLINE}-L${ESCAPE_RESET}] [${ESCAPE_UNDERLINE}-r${ESCAPE_RESET}]"
    echo
    echo -e "    ${ESCAPE_UNDERLINE}-L${ESCAPE_RESET}"
    echo    "    Do not include local ZFS sync."
    echo
    echo -e "    ${ESCAPE_UNDERLINE}-r${ESCAPE_RESET}"
    echo    "    Include remote ZFS sync."
    echo
    echo -e "    ${ESCAPE_UNDERLINE}-D${ESCAPE_RESET}"
    echo    "    Shows debug information."
    echo
    echo    "  DESCRIPTION"
    echo    "  Sync ZFS to its peers."
    echo
    echo    "  SAMPLES"
    echo    "  $0"
    echo
    exit 0
fi


if [[ ! `hostname` =~ ^vilya.*$ ]]; then
    echo "$SCRIPT_NAME: unsupported host!" >&2
    exit 1
fi


PID_FILE="/var/run/.$SCRIPT_NAME.pid"
PID_LAST=`cat $PID_FILE 2> /dev/null`
if [ -n "$PID_LAST" ]; then
    PID_ACTIVE=`ps -ax $PID_LAST | grep "^$PID_LAST" | grep "${SCRIPT_NAME}\$"`
    if [ -n "$PID_ACTIVE" ]; then
        echo "$SCRIPT_NAME: script is already running!" >&2
        exit 255
    fi
fi
echo $$ > $PID_FILE
trap "echo -ne '${ESCAPE_RESET}' ; rm $PID_FILE 2> /dev/null" EXIT SIGHUP SIGINT SIGTERM


function syncZfs {
    if [[ "$1" == "" || "$2" == "" || "$3" == "" || "$4" != "" ]]; then
        echo "$SCRIPT_NAME: invalid parameters to ZFS backup!" >&2
        return
    fi

    TITLE=$1
    SNAPSHOT_BASE=$2
    REMOTE_HOST=$3

    echo $TITLE

    #check if connection works
    ssh $SSH_OPTIONS -q $REMOTE_HOST exit
    if [ "$?" -ne "0" ]; then
        echo "$TITLE: connection to $REMOTE_HOST not available!" >&2
        return
    elif [[ DEBUG -ne 0 ]]; then
        echo -e "${ESCAPE_BLUE}$TITLE: verified connection${ESCAPE_RESET}"
    fi

    #decrypt destination
    GELI_STATUS=`ssh $SSH_OPTIONS -q $REMOTE_HOST geli status | grep disk0 | awk '{print $2}'`
    if [[ ! "$GELI_STATUS" == "ACTIVE" ]]; then
        if [[ DEBUG -ne 0 ]]; then
            echo -e "${ESCAPE_BLUE}$TITLE: mounting encrypted drives${ESCAPE_RESET}"
        fi
        mkdir -p /tmp/TmpUsb
        mount_msdosfs /dev/da1s1 /tmp/TmpUsb > /dev/null
        PASSWORD_DISK0=`cat /tmp/TmpUsb/geli_${SNAPSHOT_BASE}_disk0.pwd`
        echo $PASSWORD_DISK0 | ssh $SSH_OPTIONS -q $REMOTE_HOST "geli attach -j - /dev/label/disk0 > /dev/null"
        umount -f /tmp/TmpUsb/
        rmdir /tmp/TmpUsb
        ssh $SSH_OPTIONS -q $REMOTE_HOST /etc/rc.d/zfs onestart > /dev/null
    elif [[ DEBUG -ne 0 ]]; then
        echo -e "${ESCAPE_BLUE}$TITLE: encrypted drive already mounted${ESCAPE_RESET}"
    fi

    DATASETS=`zfs list -H -o name`
    for DATASET in $(eval echo $DATASETS); do
        if [[ DEBUG -ne 0 ]]; then
            echo -e "${ESCAPE_BLUE}$TITLE: $DATASET${ESCAPE_RESET}"
        fi

        DATASET_MOUNTPOINT=`zfs list -H -o mountpoint $DATASET`
        if [ "$DATASET_MOUNTPOINT" == "none" ]; then
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: skipping due to missing mount point${ESCAPE_RESET}"
            fi
            continue
        fi
    
        DATASET_READONLY=`zfs list -H -o readonly $DATASET`
        if [ "$DATASET_READONLY" == "on" ]; then
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: skipping due to read-only${ESCAPE_RESET}"
            fi
            continue
        fi

        TIMER_START=`date +%s`

        POOL=${DATASET%%/*}
        if [[ DEBUG -ne 0 ]]; then
            echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: located on pool $POOL${ESCAPE_RESET}"
        fi

        #create new snapshot if possible (ok if already exists)
        SNAPSHOT_CURR_NAME="Backup_${SNAPSHOT_BASE}"
        SNAPSHOT_NEXT_NAME="Backup_${SNAPSHOT_BASE}_Next"
        SNAPSHOT_PREV_NAME="Backup_${SNAPSHOT_BASE}_Previous"
        zfs snapshot ${DATASET}@$SNAPSHOT_NEXT_NAME 2> /dev/null
        if [[ DEBUG -ne 0  &&  $? -eq 0 ]]; then
            echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: created a new snapshot (@$SNAPSHOT_NEXT_NAME)${ESCAPE_RESET}"
        elif [[ DEBUG -ne 0 ]]; then
            echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: not creating a new snapshot (@$SNAPSHOT_NEXT_NAME)${ESCAPE_RESET}"
        fi

        #determine creation times for both prev and next snapshot
        SNAPSHOT_NEXT=`zfs list -H -t snapshot -o name,creation | grep "^$DATASET@$SNAPSHOT_NEXT_NAME"`
        SNAPSHOT_PREV=`zfs list -H -t snapshot -o name,creation | grep "^$DATASET@$SNAPSHOT_PREV_NAME"`
        SNAPSHOT_NEXT_CREATION=`echo $SNAPSHOT_NEXT | cut -d " " -f 2-`
        SNAPSHOT_PREV_CREATION=`echo $SNAPSHOT_PREV | cut -d " " -f 2-`

        #find data for Previous snapshot
        if [ -n "$SNAPSHOT_PREV_CREATION" ]; then
            SNAPSHOT_PREV_CREATION_SECONDS=`date -j -f '%a %b %d %H:%M %Y' "$SNAPSHOT_PREV_CREATION" +'%s'`
            SNAPSHOT_PREV_CREATION=`date -j -f '%a %b %d %H:%M %Y' "$SNAPSHOT_PREV_CREATION" +'%Y-%m-%d %H:%M'`
            SNAPSHOT_PREV_CREATION_TEXT=`date -j -f '%s' "$SNAPSHOT_PREV_CREATION_SECONDS" +'%Y%m%d'`
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: previous snapshot is at $SNAPSHOT_PREV_CREATION_TEXT${ESCAPE_RESET}"
            fi
            SNAPSHOT_PREV_EXISTS=1
        else
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: previous snapshot is not present${ESCAPE_RESET}"
            fi
            SNAPSHOT_PREV_EXISTS=0
        fi

        #find data for Next snapshot
        if [ -n "$SNAPSHOT_NEXT_CREATION" ]; then
            SNAPSHOT_NEXT_CREATION_SECONDS=`date -j -f '%a %b %d %H:%M %Y' "$SNAPSHOT_NEXT_CREATION" +'%s'`
            SNAPSHOT_NEXT_CREATION=`date -j -f '%a %b %d %H:%M %Y' "$SNAPSHOT_NEXT_CREATION" +'%Y-%m-%d %H:%M'`
            SNAPSHOT_NEXT_CREATION_TEXT=`date -j -f '%s' "$SNAPSHOT_NEXT_CREATION_SECONDS" +'%Y%m%d'`
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: next snapshot is at $SNAPSHOT_NEXT_CREATION_TEXT${ESCAPE_RESET}"
            fi
        fi

        #send dataset
        if [ $SNAPSHOT_PREV_EXISTS -ne 0 ]; then
            SNAPSHOT_DIFF_SIZE=`zfs send -npv -i $DATASET@$SNAPSHOT_PREV_NAME $DATASET@$SNAPSHOT_NEXT_NAME 2>&1 | grep "total estimated size is " | sed "s/total estimated size is //g"`
            SNAPSHOT_DIFF_AGE_SECONDS=$(( $SNAPSHOT_NEXT_CREATION_SECONDS - $SNAPSHOT_PREV_CREATION_SECONDS ))
            SNAPSHOT_DIFF_AGE_DAYS=$(( $SNAPSHOT_DIFF_AGE_SECONDS / 60 / 60 / 24 ))

            echo "* $DATASET: sending ${SNAPSHOT_DIFF_SIZE}B update ($SNAPSHOT_DIFF_AGE_DAYS days; $SNAPSHOT_PREV_CREATION to $SNAPSHOT_NEXT_CREATION)"
            zfs send -p -i $DATASET@$SNAPSHOT_PREV_NAME $DATASET@$SNAPSHOT_NEXT_NAME | ssh $SSH_OPTIONS $REMOTE_HOST zfs receive -du Backup-$POOL
        else
            SNAPSHOT_DIFF_SIZE=`zfs send -npv $DATASET@$SNAPSHOT_NEXT_NAME 2>&1 | grep "total estimated size is " | sed "s/total estimated size is //g"`

            echo "* $DATASET: sending initial ${SNAPSHOT_DIFF_SIZE}B (all to $SNAPSHOT_NEXT_CREATION)"
            zfs send -p $DATASET@$SNAPSHOT_NEXT_NAME | ssh $SSH_OPTIONS $REMOTE_HOST zfs receive -du Backup-$POOL

            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: removing mounting point at destination${ESCAPE_RESET}"
            fi
            ssh $SSH_OPTIONS $REMOTE_HOST zfs set mountpoint=none Backup-$DATASET
        fi

        if [ "$?" -eq "0" ]; then
            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: destination snapshot cleanup"
            fi
            ssh $SSH_OPTIONS $REMOTE_HOST zfs destroy Backup-$DATASET@$SNAPSHOT_PREV_NAME 2> /dev/null
            ssh $SSH_OPTIONS $REMOTE_HOST zfs rename Backup-$DATASET@$SNAPSHOT_NEXT_NAME $SNAPSHOT_CURR_NAME 2> /dev/null
            if [[ DEBUG -ne 0 ]]; then
                echo -ne "${ESCAPE_RESET}"
            fi

            if [[ DEBUG -ne 0 ]]; then
                echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: source snapshot cleanup"
            fi
            if [ $SNAPSHOT_PREV_EXISTS -ne 0 ]; then
                zfs destroy $DATASET@$SNAPSHOT_PREV_NAME
            fi

            zfs rename $DATASET@$SNAPSHOT_NEXT_NAME $SNAPSHOT_PREV_NAME

            if [[ DEBUG -ne 0 ]]; then
                echo -ne "${ESCAPE_RESET}"
            fi
        else
            echo "$TITLE: $DATASET: cannot send snapshot!" >&2
            HAD_ERRORS=1
        fi

        TIMER_END=`date +%s`
        if [[ DEBUG -ne 0 ]]; then
            TIMER_DURATION=$(( $TIMER_END - $TIMER_START ))
            TIMER_DURATION_MINUTES=$(( $TIMER_DURATION / 60 ))
            TIMER_DURATION_SECONDS=$(( $TIMER_DURATION % 60 ))
            echo -e "${ESCAPE_BLUE}$TITLE: $DATASET: operation completed in $TIMER_DURATION_MINUTES minutes and $TIMER_DURATION_SECONDS seconds${ESCAPE_RESET}"
        fi
    done
}


if [[ $LOCAL -ne 0 ]]; then
    syncZfs 'ZFS Local' 'LocalBackup' '192.168.1.2'
fi

if [[ $REMOTE -ne 0 ]]; then
    syncZfs 'ZFS Remote' 'RemoteBackup' '192.168.2.2'
fi

